@@ -1,12 +1,10 @@
 #include "cheapr.h"
-#include <vector>
 
 // static cpp11::writable::integers CHEAPR_ZERO(1);
 // void constants_init(DllInfo* dll){
 //   CHEAPR_ZERO[0] = 0;
 // }
 
-[[cpp11::register]]
 SEXP rebuild(SEXP x, SEXP source, bool shallow_copy){
   if (is_df(x)){
     if (is_bare_df(source)){
@@ -51,7 +49,7 @@ SEXP cpp_rep_len(SEXP x, int length){
     if (out_size == df_nrow(x)) return x;
     int n_cols = Rf_length(x);
     SEXP out = SHIELD(new_vec(VECSXP, n_cols));
-    const SEXP *p_x = VECTOR_PTR_RO(x);
+    const SEXP *p_x = LIST_PTR_RO(x);
     for (int i = 0; i < n_cols; ++i){
       SET_VECTOR_ELT(out, i, cpp_rep_len(p_x[i], length));
     }
@@ -62,7 +60,7 @@ SEXP cpp_rep_len(SEXP x, int length){
     SHIELD(out = rebuild(out, x, false));
     YIELD(3);
     return out;
-  } else if (is_simple_vec2(x)){
+  } else if (cheapr_is_simple_vec2(x)){
 
     int size = Rf_length(x);
     int n_chunks, k, chunk_size;
@@ -95,7 +93,7 @@ SEXP cpp_rep_len(SEXP x, int length){
       return out;
     }
     case CHEAPR_INT64SXP: {
-      const int64_t *p_x = INTEGER64_RO_PTR(x);
+      const int64_t *p_x = INTEGER64_PTR_RO(x);
       SEXP out = SHIELD(new_vec(REALSXP, out_size));
       int64_t* RESTRICT p_out = INTEGER64_PTR(out);
 
@@ -184,7 +182,7 @@ SEXP cpp_rep_len(SEXP x, int length){
       return out;
     }
     case VECSXP: {
-      const SEXP *p_x = VECTOR_PTR_RO(x);
+      const SEXP *p_x = LIST_PTR_RO(x);
       SEXP out = SHIELD(new_vec(VECSXP, out_size));
 
       if (size == 1){
@@ -226,11 +224,11 @@ SEXP cpp_rep_len(SEXP x, int length){
 [[cpp11::register]]
 SEXP cpp_rep(SEXP x, SEXP times){
 
-  R_xlen_t n = vec_length(x);
+  R_xlen_t n = vector_length(x);
 
   R_xlen_t out_size;
   R_xlen_t n_times = Rf_xlength(times);
-  SHIELD(times = coerce_vector(times, INTSXP));
+  SHIELD(times = cast<r_integer_t>(times, R_NilValue));
 
   if (n_times == 1){
     out_size = n * INTEGER(times)[0];
@@ -262,7 +260,7 @@ SEXP cpp_rep(SEXP x, SEXP times){
         return out;
       }
 
-    } else if (is_simple_vec2(x)){
+    } else if (cheapr_is_simple_vec2(x)){
       SEXP out = SHIELD(new_vec(TYPEOF(x), cpp_sum(times)));
       switch (TYPEOF(x)){
       case LGLSXP:
@@ -309,7 +307,7 @@ SEXP cpp_rep(SEXP x, SEXP times){
         return out;
       }
       case VECSXP: {
-        const SEXP *p_x = VECTOR_PTR_RO(x);
+        const SEXP *p_x = LIST_PTR_RO(x);
         R_xlen_t k = 0;
         for (R_xlen_t i = 0; i < n; ++i){
           for (int j = 0; j < p_times[i]; ++j, ++k) SET_VECTOR_ELT(out, k, p_x[i]);
@@ -335,82 +333,75 @@ SEXP cpp_rep(SEXP x, SEXP times){
 [[cpp11::register]]
 SEXP cpp_rep_each(SEXP x, SEXP each){
   int32_t NP = 0;
-  SHIELD(each = coerce_vector(each, INTSXP)); ++NP;
+  SHIELD(each = cast<r_integer_t>(each, R_NilValue)); ++NP;
   if (Rf_length(each) == 1){
     if (INTEGER(each)[0] == 1){
-      Rf_unprotect(NP);
+      YIELD(NP);
       return x;
     }
-    SHIELD(each = cpp_rep_len(each, vec_length(x))); ++NP;
+    SHIELD(each = cpp_rep_len(each, vector_length(x))); ++NP;
   }
   SEXP out = SHIELD(cpp_rep(x, each)); ++NP;
   YIELD(NP);
   return out;
 }
 
+R_xlen_t length_common(SEXP x){
+
+  if (TYPEOF(x) != VECSXP){
+    Rf_error("x` must be a list");
+  }
+
+  R_xlen_t n = Rf_xlength(x);
+
+  const SEXP *p_x = LIST_PTR_RO(x);
+  R_xlen_t out = 0;
+
+  for (R_xlen_t i = 0; i < n; ++i){
+    if (is_null(p_x[i])) continue;
+
+    if (vector_length(p_x[i]) == 0){
+      out = 0;
+      break;
+    }
+    out = std::max(out, vector_length(p_x[i]));
+  }
+  return out;
+}
+
 // Recycle elements of a list `x`
 
+void recycle_in_place(SEXP x, R_xlen_t n){
+
+  int xn = Rf_length(x);
+  const SEXP *p_x = LIST_PTR_RO(x);
+
+  for (int i = 0; i < xn; ++i){
+    if (!is_null(p_x[i])){
+      SET_VECTOR_ELT(x, i, cpp_rep_len(p_x[i], n));
+    }
+  }
+}
+
 [[cpp11::register]]
 SEXP cpp_recycle(SEXP x, SEXP length){
 
-  int32_t NP = 0;
-  int xn = Rf_length(x);
-  int n_objs = xn - static_cast<int>(null_count(x));
-
-  const SEXP *p_x = VECTOR_PTR_RO(x);
-  SEXP out = SHIELD(new_vec(VECSXP, n_objs)); ++NP;
-  SEXP sizes = SHIELD(cpp_lengths(x, false)); ++NP;
-  const int* RESTRICT p_sizes = INTEGER_RO(sizes);
-  bool has_length = !is_null(length);
-  SHIELD(length = coerce_vec(length, INTSXP)); ++NP;
+  SEXP out = SHIELD(cpp_drop_null(x, true));
 
   int n = 0;
 
-  if (!has_length){
-    if (n_objs > 0){
-      n = -1;
-      // We calculate `max(sizes)`
-      // We won't have any NA in sizes so no NA checking is needed
-      for (int i = 0; i < xn; ++i){
-        if (is_null(p_x[i])){
-          continue;
-        } else if (p_sizes[i] == 0){
-          n = 0;
-          break;
-        } else {
-          n = std::max(n, p_sizes[i]);
-        }
-      }
-    }
-  } else {
+  if (!is_null(length)){
     n = Rf_asInteger(length);
-  }
-
-  n = std::max(n, 0);
-
-  int k = 0;
-
-  SEXP names = SHIELD(get_names(x)); ++NP;
-  SEXP new_names = R_NilValue;
-
-  if (xn == n_objs || is_null(names)){
-    for (int i = 0; i < xn; ++i){
-      if (is_null(p_x[i])) continue;
-      SET_VECTOR_ELT(out, k++, cpp_rep_len(p_x[i], n));
+    if (n < 0){
+      Rf_error("Recycled `length` must be >= 0 in %s", __func__);
     }
-    set_names(out, names);
   } else {
-    new_names = SHIELD(new_vec(STRSXP, n_objs)); ++NP;
-    const SEXP *p_names = STRING_PTR_RO(names);
-    for (int i = 0; i < xn; ++i){
-      if (is_null(p_x[i])) continue;
-      SET_VECTOR_ELT(out, k, cpp_rep_len(p_x[i], n));
-      SET_STRING_ELT(new_names, k++, p_names[i]);
-    }
-    set_names(out, new_names);
+    n = length_common(out);
   }
 
-  YIELD(NP);
+  recycle_in_place(out, n);
+
+  YIELD(1);
   return out;
 }
 
@@ -420,7 +411,7 @@ SEXP cpp_unique(SEXP x, bool names){
 
   int32_t NP = 0;
 
-  bool is_simple = is_simple_atomic_vec(x);
+  bool is_simple = cheapr_is_simple_atomic_vec(x);
 
   if (is_compact_seq(x)){
     return x;
@@ -472,7 +463,7 @@ SEXP cpp_setdiff(SEXP x, SEXP y, bool unique){
   } else {
     SHIELD(x);
   }
-  SEXP matches = SHIELD(Rf_match(y, x, NA_INTEGER));
+  SEXP matches = SHIELD(match(y, x, NA_INTEGER));
   SEXP locs = SHIELD(cpp_which_na(matches));
   if (Rf_xlength(locs) == Rf_xlength(x)){
     YIELD(3);
@@ -492,7 +483,7 @@ SEXP cpp_intersect(SEXP x, SEXP y, bool unique){
   } else {
     SHIELD(x);
   }
-  SEXP matches = SHIELD(Rf_match(y, x, NA_INTEGER));
+  SEXP matches = SHIELD(match(y, x, NA_INTEGER));
   SEXP locs = SHIELD(cpp_which_not_na(matches));
   if (Rf_xlength(locs) == Rf_xlength(x)){
     YIELD(3);
@@ -504,38 +495,6 @@ SEXP cpp_intersect(SEXP x, SEXP y, bool unique){
     return out;
   }
 }
-// SEXP cpp_setdiff(SEXP x, SEXP y, bool dups){
-//
-//   bool is_simple = is_simple_atomic_vec(x) && is_simple_atomic_vec(y);
-//
-//   int32_t NP = 0;
-//   if (is_simple){
-//     if (!dups){
-//       SHIELD(x = cpp_unique(x)); ++NP;
-//     }
-//     SEXP matches;
-//     if (Rf_length(x) < 10000){
-//       matches = SHIELD(Rf_match(y, x, NA_INTEGER)); ++NP;
-//     } else {
-//       matches = SHIELD(cheapr_fast_match(x, y)); ++NP;
-//     }
-//     SEXP locs = SHIELD(cpp_which_na(matches)); ++NP;
-//     SEXP out = SHIELD(sset_vec(x, locs, false)); ++NP;
-//
-//     Rf_copyMostAttrib(x, out);
-//     YIELD(NP);
-//     return out;
-//   } else {
-//     if (!dups){
-//       SHIELD(x = cheapr_fast_unique(x)); ++NP;
-//     }
-//     SEXP matches = SHIELD(cheapr_fast_match(x, y)); ++NP;
-//     SEXP locs = SHIELD(cpp_which_na(matches)); ++NP;
-//     SEXP out = SHIELD(cheapr_sset(x, locs)); ++NP;
-//     YIELD(NP);
-//     return out;
-//   }
-// }
 
 [[cpp11::register]]
 SEXP cpp_na_init(SEXP x, int n){
@@ -566,69 +525,25 @@ SEXP get_ptypes(SEXP x){
   return out;
 }
 
-SEXP cpp_factor_as_character(SEXP x){
+// Helper to turn a factor into a character vec
+SEXP factor_as_character(SEXP x){
   return sset_vec(Rf_getAttrib(x, R_LevelsSymbol), x, true);
 }
 
-[[cpp11::register]]
-SEXP cpp_combine_levels(SEXP x){
-  if (TYPEOF(x) != VECSXP){
-    Rf_error("`x` must be a list of factors in %s", __func__);
-  }
-  int n = Rf_length(x);
-  SEXP levels = SHIELD(new_vec(VECSXP, n));
-  const SEXP *p_x = VECTOR_PTR_RO(x);
+// Helper to turn a character vec into a factor, given levels
+SEXP character_as_factor(SEXP x, SEXP levels){
 
-  SEXP fct_levels;
-
-  PROTECT_INDEX fct_idx;
-  R_ProtectWithIndex(fct_levels = R_NilValue, &fct_idx);
-
-  for (int i = 0; i < n; ++i){
-    if (Rf_isFactor(p_x[i])){
-      fct_levels = Rf_getAttrib(p_x[i], R_LevelsSymbol);
-    } else {
-      R_Reprotect(fct_levels = base_as_character(p_x[i]), fct_idx);
-    }
-    SET_VECTOR_ELT(levels, i, fct_levels);
-  }
-  SEXP out = SHIELD(cpp_c(levels));
-  SHIELD(out = cpp_unique(out, false));
-  YIELD(4);
-  return out;
-}
-
-SEXP cpp_combine_factors(SEXP x){
-
-  if (TYPEOF(x) != VECSXP){
-    Rf_error("`x` must be a list of factors in %s", __func__);
+  if (TYPEOF(x) != STRSXP){
+    Rf_error("`x` must be a character vector in %s", __func__);
   }
 
-  int n = Rf_length(x);
-  const SEXP *p_x = VECTOR_PTR_RO(x);
+  int32_t NP = 0;
 
-  SEXP levels = SHIELD(cpp_combine_levels(x));
-  SEXP chars = SHIELD(new_vec(VECSXP, n));
-  SEXP char_vec;
-
-  PROTECT_INDEX char_vec_idx;
-  R_ProtectWithIndex(char_vec = R_NilValue, &char_vec_idx);
-
-  for (int i = 0; i < n; ++i){
-    if (Rf_isFactor(p_x[i])){
-      R_Reprotect(char_vec = cpp_factor_as_character(p_x[i]), char_vec_idx);
-    } else {
-      R_Reprotect(char_vec = base_as_character(p_x[i]), char_vec_idx);
-    }
-    SET_VECTOR_ELT(chars, i, char_vec);
-  }
-
-  R_Reprotect(char_vec = cpp_c(chars), char_vec_idx);
-  SEXP out = SHIELD(cheapr_factor(char_vec,
-    cpp11::named_arg("levels") = levels,
-    cpp11::named_arg("na_exclude") = !cpp_any_na(levels, false)
-  ));
-  YIELD(4);
+  SEXP out = SHIELD(match(levels, x, NA_INTEGER)); ++NP;
+  SEXP cls = SHIELD(make_utf8_str("factor")); ++NP;
+  Rf_setAttrib(out, R_LevelsSymbol, levels);
+  Rf_classgets(out, cls);
+  YIELD(NP);
   return out;
 }
 
@@ -639,7 +554,7 @@ SEXP cpp_combine_factors(SEXP x){
 SEXP cpp_list_c(SEXP x){
   int32_t NP = 0;
   R_xlen_t n = Rf_xlength(x);
-  const SEXP *p_x = VECTOR_PTR_RO(x);
+  const SEXP *p_x = LIST_PTR_RO(x);
 
   R_xlen_t out_size = 0;
   for (R_xlen_t i = 0; i < n; ++i){
@@ -669,17 +584,17 @@ SEXP cpp_list_c(SEXP x){
     const SEXP *p_temp;
 
     if (TYPEOF(p_x[i]) == VECSXP){
-      p_temp = VECTOR_PTR_RO(p_x[i]);
+      p_temp = LIST_PTR_RO(p_x[i]);
       R_Reprotect(names = get_names(p_x[i]), nm_idx);
       m = Rf_xlength(p_x[i]);
     } else {
       SET_VECTOR_ELT(container_list, 0, p_x[i]);
       if (x_has_names){
-        R_Reprotect(names = Rf_ScalarString(STRING_ELT(x_names, i)), nm_idx);
+        R_Reprotect(names = as_r_scalar(STRING_ELT(x_names, i)), nm_idx);
       } else {
         R_Reprotect(names = R_NilValue, nm_idx);
       }
-      p_temp = VECTOR_PTR_RO(container_list);
+      p_temp = LIST_PTR_RO(container_list);
       m = 1;
     }
 
@@ -702,32 +617,8 @@ SEXP cpp_list_c(SEXP x){
   return out;
 }
 
-// Helper to concatenate 2 lists
-
-SEXP list_c2(SEXP x, SEXP y){
-  SEXP temp = SHIELD(new_vec(VECSXP, 2));
-
-  SET_VECTOR_ELT(temp, 0, x);
-  SET_VECTOR_ELT(temp, 1, y);
-  SEXP out = SHIELD(cpp_list_c(temp));
-
-  YIELD(2);
-  return out;
-}
-
-// Helper to concatenate two vectors
-SEXP c2(SEXP x, SEXP y){
-  SEXP temp = SHIELD(new_vec(VECSXP, 2));
-
-  SET_VECTOR_ELT(temp, 0, x);
-  SET_VECTOR_ELT(temp, 1, y);
-  SEXP out = SHIELD(cpp_c(temp));
-
-  YIELD(2);
-  return out;
-}
-
 // Concatenate a list of data frames
+
 SEXP cpp_df_c(SEXP x){
 
   if (TYPEOF(x) != VECSXP){
@@ -739,100 +630,99 @@ SEXP cpp_df_c(SEXP x){
   if (n_frames == 0) return R_NilValue;
 
   int32_t NP = 0;
-  const SEXP *p_x = VECTOR_PTR_RO(x);
 
-  SEXP df = p_x[0];
+  // Since we are casting potential non-dfs to dfs we need a new list
+  SEXP frames = SHIELD(init<r_list_t>(n_frames, false)); ++NP;
 
-  SEXP names;
-  PROTECT_INDEX names_idx;
-  R_ProtectWithIndex(names = get_names(df), &names_idx); ++NP;
+  const SEXP *p_x = LIST_PTR_RO(x);
+  const SEXP *p_frames = LIST_PTR_RO(frames);
 
-  if (!is_df(df)){
-    YIELD(NP); Rf_error("Can't combine data frames with non data frames");
-  }
+  SEXP df;
+  PROTECT_INDEX df_idx;
+  R_ProtectWithIndex(df = cast<r_data_frame_t>(p_x[0], R_NilValue), &df_idx); ++NP;
+  SET_VECTOR_ELT(frames, 0, df);
 
-  SEXP df_template = df;
+  SEXP new_names, df_names, ptype_names;
+  PROTECT_INDEX new_names_idx, df_names_idx, ptype_names_idx;
 
-  SEXP frames = SHIELD(new_vec(VECSXP, n_frames)); ++NP;
-  SET_VECTOR_ELT(frames, 0, df_template);
-
-  SEXP ptypes, new_names, new_ptypes, new_cols,
-  temp_list, df_names, ptype_names;
-  PROTECT_INDEX
-  new_names_idx, ptypes_idx, new_ptypes_idx, new_cols_idx,
-  temp_list_idx, df_names_idx, ptype_names_idx;
-
-  R_ProtectWithIndex(ptypes = get_ptypes(df), &ptypes_idx); ++NP;
   R_ProtectWithIndex(new_names = R_NilValue, &new_names_idx); ++NP;
-  R_ProtectWithIndex(new_ptypes = R_NilValue, &new_ptypes_idx); ++NP;
-  R_ProtectWithIndex(new_cols = R_NilValue, &new_cols_idx); ++NP;
-  R_ProtectWithIndex(temp_list = new_vec(VECSXP, 2), &temp_list_idx); ++NP;
   R_ProtectWithIndex(df_names = R_NilValue, &df_names_idx); ++NP;
-  R_ProtectWithIndex(ptype_names = R_NilValue, &ptype_names_idx); ++NP;
+  R_ProtectWithIndex(ptype_names = get_names(df), &ptype_names_idx); ++NP;
+
+  int n_cols = Rf_length(ptype_names);
 
   // We do 2 passes
-  // 1st pass: Check inputs are dfs and construct prototype list
+  // 1st pass: Cast inputs to dfs and construct prototypes
   // 2nd pass: initialise data frame vecs (if need be) and combine simultaneously
 
   int out_size = df_nrow(df);
 
+  bool na_padding = false;
+
   for (int i = 1; i < n_frames; ++i){
-    df = p_x[i];
-
-    if (!is_df(df)){
-      YIELD(NP); Rf_error("Can't combine data frames with non data frames");
-    }
-
+    R_Reprotect(df = cast<r_data_frame_t>(p_x[i], R_NilValue), df_idx);
     R_Reprotect(df_names = get_names(df), df_names_idx);
-    R_Reprotect(ptype_names = get_names(ptypes), ptype_names_idx);
 
     R_Reprotect(new_names = cpp_setdiff(
       df_names, ptype_names, false
     ), new_names_idx);
 
-    // Adjust prototype list
+    // Adjust prototype names
     if (Rf_length(new_names) > 0){
-      R_Reprotect(new_cols = cpp_df_select(df, new_names), new_cols_idx);
-      R_Reprotect(new_ptypes = get_ptypes(new_cols), new_ptypes_idx);
-      SET_VECTOR_ELT(temp_list, 0, ptypes);
-      SET_VECTOR_ELT(temp_list, 1, new_ptypes);
-      R_Reprotect(ptypes = cpp_list_c(temp_list), ptypes_idx);
-      SET_VECTOR_ELT(temp_list, 0, names);
-      SET_VECTOR_ELT(temp_list, 1, new_names);
-      R_Reprotect(names = cpp_c(temp_list), names_idx);
-      set_names(ptypes, names);
+      na_padding = true;
+      R_Reprotect(ptype_names = r_combine(ptype_names, new_names), ptype_names_idx);
     }
+    na_padding = na_padding || Rf_length(df) != n_cols;
     out_size += df_nrow(df);
+    SET_VECTOR_ELT(frames, i, df);
   }
 
-  int n_cols = Rf_length(names);
+  n_cols = Rf_length(ptype_names);
 
-  SEXP vec, name_locs;
-  PROTECT_INDEX vec_idx, name_locs_idx;
+  SEXP vec;
+  PROTECT_INDEX vec_idx;
   R_ProtectWithIndex(vec = R_NilValue, &vec_idx); ++NP;
-  R_ProtectWithIndex(name_locs = R_NilValue, &name_locs_idx); ++NP;
 
   SEXP out = SHIELD(new_vec(VECSXP, n_cols)); ++NP;
   SEXP vectors = SHIELD(new_vec(VECSXP, n_frames)); ++NP;
 
-  const SEXP *p_ptypes = VECTOR_PTR_RO(ptypes);
-  const SEXP *p_names = VECTOR_PTR_RO(names);
+  const SEXP *p_ptype_names = STRING_PTR_RO(ptype_names);
 
-  for (int j = 0; j < n_cols; ++j){
-    for (int i = 0; i < n_frames; ++i){
-      vec = get_list_element(p_x[i], p_names[j]);
-
-      if (is_null(vec)){
-        R_Reprotect(vec = cpp_na_init(p_ptypes[j], df_nrow(p_x[i])), vec_idx);
+  if (na_padding){
+    // Get archetype of each col
+    SEXP vec_archetypes = SHIELD(new_vec(VECSXP, n_cols)); ++NP;
+    const SEXP *p_vec_archetypes = LIST_PTR_RO(vec_archetypes);
+    for (int j = 0; j < n_cols; ++j){
+      for (int i = 0; i < n_frames; ++i){
+        SET_VECTOR_ELT(vectors, i, get_list_element(p_frames[i], p_ptype_names[j]));
       }
-      SET_VECTOR_ELT(vectors, i, vec);
+      SET_VECTOR_ELT(vec_archetypes, j, cpp_common_template(vectors));
+    }
+    // Combine all cols
+    for (int j = 0; j < n_cols; ++j){
+      for (int i = 0; i < n_frames; ++i){
+        vec = get_list_element(p_frames[i], p_ptype_names[j]);
+        if (is_null(vec)){
+          R_Reprotect(vec = cpp_na_init(p_vec_archetypes[j], df_nrow(p_frames[i])), vec_idx);
+        }
+        SET_VECTOR_ELT(vectors, i, vec);
+      }
+      SET_VECTOR_ELT(out, j, combine_internal(vectors, out_size, p_vec_archetypes[j]));
+    }
+  } else {
+    // Simply combine rows with no NA padding
+    for (int j = 0; j < n_cols; ++j){
+      for (int i = 0; i < n_frames; ++i){
+        SET_VECTOR_ELT(vectors, i, get_list_element(p_frames[i], p_ptype_names[j]));
+      }
+      SET_VECTOR_ELT(out, j, cpp_c(vectors));
     }
-    SET_VECTOR_ELT(out, j, cpp_c(vectors));
   }
+
   set_list_as_df(out);
   Rf_setAttrib(out, R_RowNamesSymbol, create_df_row_names(out_size));
-  set_names(out, names);
-  SHIELD(out = rebuild(out, df_template, false)); ++NP;
+  set_names(out, ptype_names);
+  SHIELD(out = rebuild(out, VECTOR_ELT(frames, 0), false)); ++NP;
   YIELD(NP);
   return out;
 }
@@ -850,7 +740,7 @@ SEXP cpp_df_col_c(SEXP x, bool recycle, bool name_repair){
   }
 
   int n = Rf_length(x);
-  const SEXP *p_x = VECTOR_PTR_RO(x);
+  const SEXP *p_x = LIST_PTR_RO(x);
 
   int out_ncols = 0;
 
@@ -861,10 +751,10 @@ SEXP cpp_df_col_c(SEXP x, bool recycle, bool name_repair){
 
   for (int i = 0; i < n; ++i){
     if (is_df(p_x[i])){
-      df_pointers[i] = VECTOR_PTR_RO(p_x[i]);
+      df_pointers[i] = LIST_PTR_RO(p_x[i]);
       out_ncols += Rf_length(p_x[i]);
     } else {
-      df_pointers[i] = VECTOR_PTR_RO(container_list);
+      df_pointers[i] = LIST_PTR_RO(container_list);
       ++out_ncols;
     }
   }
@@ -894,7 +784,7 @@ SEXP cpp_df_col_c(SEXP x, bool recycle, bool name_repair){
     } else {
       SET_VECTOR_ELT(container_list, 0, p_x[i]);
       if (x_has_names){
-        R_Reprotect(names = Rf_ScalarString(STRING_ELT(x_names, i)), nm_idx);
+        R_Reprotect(names = as_r_scalar(STRING_ELT(x_names, i)), nm_idx);
       } else {
         R_Reprotect(names = R_NilValue, nm_idx);
       }
@@ -919,7 +809,7 @@ SEXP cpp_df_col_c(SEXP x, bool recycle, bool name_repair){
 
   SEXP r_nrows = SHIELD(R_NilValue); ++NP;
   if (Rf_length(out) == 0 && Rf_length(x) != 0){
-    SHIELD(r_nrows = Rf_ScalarInteger(vec_length(VECTOR_ELT(x, 0)))); ++NP;
+    SHIELD(r_nrows = as_r_scalar<int>(vector_length(VECTOR_ELT(x, 0)))); ++NP;
   }
 
   SHIELD(out = cpp_new_df(out, r_nrows, false, name_repair)); ++NP;
@@ -931,177 +821,272 @@ SEXP cpp_df_col_c(SEXP x, bool recycle, bool name_repair){
   return out;
 }
 
-// `c()` but no concatenation of names
-[[cpp11::register]]
-SEXP cpp_c(SEXP x){
-  if (TYPEOF(x) != VECSXP){
-    Rf_error("`x` must be a list of vectors");
+// define CHEAPR_COMBINE_NAMES
+// if (any_names){
+//   R_xlen_t ii = 0;
+//   SHIELD(combined_names = new_vec(STRSXP, out_size)); ++NP;
+//   for (int i = 0; i < n; ++i){
+//     R_Reprotect(vec_names = get_vec_names(p_x[i]), vec_names_idx);
+//     const SEXP *p_vec_names = STRING_PTR_RO(vec_names);
+//     if (is_null(vec_names)){
+//       ii += vector_length(p_x[i]);
+//     } else {
+//       for (R_xlen_t j = 0; j < Rf_xlength(vec_names); ++j, ++ii){
+//         SET_STRING_ELT(combined_names, ii, p_vec_names[j]);
+//       }
+//     }
+//   }
+// }
+
+// Combine vectors given the following args:
+// * A list of vectors
+// * The final size of the combined vector
+// * A template vector (of length 0) whose type is the common type between all vectors
+// This allows us to eliminate a few loops
+SEXP combine_internal(SEXP x, const R_xlen_t out_size, SEXP vec_template){
+
+  if (vector_length(vec_template) != 0){
+    Rf_error("`vec_template` must be of length 0");
   }
-  int32_t NP = 0;
+
   int n = Rf_length(x);
-  const SEXP *p_x = VECTOR_PTR_RO(x);
-  if (n == 1){
-    return p_x[0];
-  }
+  const SEXP *p_x = LIST_PTR_RO(x);
 
-  int vector_type = NILSXP;
-  R_xlen_t out_size = 0;
-
-  bool is_factor = false, is_frame = false, is_classed = false,
-    is_simple = false, is_date = false, is_datetime = false;
-
-
-  // We use the tz info of the first datetime vec to copy to final result
-  int first_datetime = INTEGER_MAX;
-
-  SEXP elem = R_NilValue;
-
-  for (int i = 0; i < n; ++i){
-    elem = p_x[i];
-    vector_type = std::max(vector_type, TYPEOF(elem));
-    out_size += Rf_xlength(elem);
-    is_factor = is_factor || Rf_isFactor(elem);
-    is_simple = is_simple || is_simple_atomic_vec(elem);
-    is_date = is_date || Rf_inherits(elem, "Date");
-    is_datetime = is_datetime || Rf_inherits(elem, "POSIXct");
-    if (is_datetime){
-      first_datetime = std::min(i, first_datetime);
-    }
-    is_frame = is_frame || is_df(elem);
-    is_classed = is_classed || Rf_isObject(elem);
-  }
-
-  // Date vectors can be ints but datetimes can't
-  bool is_date2 = is_date && !is_datetime && (vector_type == INTSXP || vector_type == REALSXP);
-  bool is_datetime2 = !is_date && is_datetime && vector_type == REALSXP;
-
-  if (is_frame){
-    return cpp_df_c(x);
-  }
-
-  if (is_factor){
-    return cpp_combine_factors(x);
-  }
-
-  if (is_classed && !(is_date2 || is_datetime2)){
-    SEXP call = SHIELD(coerce_vec(x, LISTSXP));
-    SHIELD(call = Rf_lcons(install_utf8("c"), call));
-    SEXP out = SHIELD(Rf_eval(call, R_GetCurrentEnv()));
-    YIELD(3);
-    return out;
-  }
+  int32_t NP = 0;
 
   R_xlen_t k = 0;
   R_xlen_t m = 0;
-  SEXP out, temp;
+  SEXP out = R_NilValue;
 
-  PROTECT_INDEX temp_idx;
-  R_ProtectWithIndex(temp = R_NilValue, &temp_idx); ++NP;
+  SEXP vec;
+  PROTECT_INDEX vec_idx;
+  R_ProtectWithIndex(vec = R_NilValue, &vec_idx); ++NP;
 
+  r_type common = get_r_type(vec_template);
+  SEXP combined_names = R_NilValue;
+  bool has_top_level_names = vec_has_names(x);
 
-  switch(vector_type){
-  case NILSXP: {
-    out = SHIELD(R_NilValue); ++NP;
+  // If in the future it is decided that inner names are to be kept
+  // Use this condition within the inner loop and then use the macro
+  // defined above combine_internal
+  // any_names = !has_top_level_names && (any_names || has_names(vec));
+
+  switch (common){
+  case r_null: {
     break;
   }
-  case LGLSXP:
-  case INTSXP: {
+  case r_lgl: {
+
+    SHIELD(out = init<r_logical_t>(out_size, false)); ++NP;
 
-    out = SHIELD(new_vec(vector_type, out_size)); ++NP;
     int* RESTRICT p_out = INTEGER(out);
 
     for (int i = 0; i < n; ++i, k += m){
-      if (TYPEOF(p_x[i]) == vector_type){
-        temp = p_x[i];
-      } else {
-        R_Reprotect(temp = coerce_vec(p_x[i], vector_type), temp_idx);
-      }
-      m = Rf_xlength(temp);
-      std::copy_n(INTEGER_RO(temp), m, &p_out[k]);
+      R_Reprotect(vec = cast<r_logical_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(INTEGER_RO(vec), m, &p_out[k]);
     }
     break;
   }
-  case REALSXP: {
+  case r_int: {
+
+    SHIELD(out = init<r_integer_t>(out_size, false)); ++NP;
+
+    int* RESTRICT p_out = INTEGER(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_integer_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(INTEGER_RO(vec), m, &p_out[k]);
+    }
+    break;
+  }
+  case r_int64: {
+
+    SHIELD(out = init<r_integer64_t>(out_size, false)); ++NP;
+
+    int64_t* RESTRICT p_out = INTEGER64_PTR(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_integer64_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(INTEGER64_PTR_RO(vec), m, &p_out[k]);
+    }
+    break;
+  }
+  case r_dbl: {
+
+    SHIELD(out = init<r_numeric_t>(out_size, false)); ++NP;
 
-    out = SHIELD(new_vec(vector_type, out_size)); ++NP;
     double* RESTRICT p_out = REAL(out);
 
     for (int i = 0; i < n; ++i, k += m){
-      if (TYPEOF(p_x[i]) == vector_type){
-        temp = p_x[i];
-      } else {
-        R_Reprotect(temp = coerce_vec(p_x[i], vector_type), temp_idx);
-      }
-      m = Rf_xlength(temp);
-      std::copy_n(REAL_RO(temp), m, &p_out[k]);
+      R_Reprotect(vec = cast<r_numeric_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(REAL_RO(vec), m, &p_out[k]);
     }
     break;
   }
-  case STRSXP: {
+  case r_chr: {
 
-    out = SHIELD(new_vec(vector_type, out_size)); ++NP;
+    SHIELD(out = init<r_character_t>(out_size, false)); ++NP;
 
     for (int i = 0; i < n; ++i){
-      if (TYPEOF(p_x[i]) == vector_type){
-        temp = p_x[i];
-      } else {
-        R_Reprotect(temp = coerce_vec(p_x[i], vector_type), temp_idx);
-      }
-      m = Rf_xlength(temp);
-      const SEXP *p_temp = STRING_PTR_RO(temp);
+      R_Reprotect(vec = cast<r_character_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+
+      const SEXP *p_vec = STRING_PTR_RO(vec);
       for (R_xlen_t j = 0; j < m; ++k, ++j){
-        SET_STRING_ELT(out, k, p_temp[j]);
+        SET_STRING_ELT(out, k, p_vec[j]);
       }
     }
     break;
   }
-  case CPLXSXP: {
+  case r_cplx: {
+
+    SHIELD(out = init<r_complex_t>(out_size, false)); ++NP;
 
-    out = SHIELD(new_vec(vector_type, out_size)); ++NP;
     Rcomplex *p_out = COMPLEX(out);
 
     for (int i = 0; i < n; ++i, k += m){
-      if (TYPEOF(p_x[i]) == vector_type){
-        temp = p_x[i];
-      } else {
-        R_Reprotect(temp = coerce_vec(p_x[i], vector_type), temp_idx);
-      }
-      m = Rf_xlength(temp);
-      std::copy_n(COMPLEX_RO(temp), m, p_out + k);
+      R_Reprotect(vec = cast<r_complex_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(COMPLEX_RO(vec), m, &p_out[k]);
     }
     break;
   }
-  case VECSXP: {
+  case r_raw: {
 
-    out = SHIELD(new_vec(vector_type, out_size)); ++NP;
+    SHIELD(out = init<r_raw_t>(out_size, false)); ++NP;
+
+    Rbyte *p_out = RAW(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_raw_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(RAW_RO(vec), m, &p_out[k]);
+    }
+    break;
+  }
+  case r_list: {
+
+    SHIELD(out = init<r_list_t>(out_size, false)); ++NP;
 
     for (int i = 0; i < n; ++i){
-      if (TYPEOF(p_x[i]) == vector_type){
-        temp = p_x[i];
-      } else {
-        R_Reprotect(temp = coerce_vec(p_x[i], vector_type), temp_idx);
-      }
-      m = Rf_xlength(temp);
-      const SEXP *p_temp = VECTOR_PTR_RO(temp);
+      R_Reprotect(vec = cast<r_list_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+
+      const SEXP *p_vec = LIST_PTR_RO(vec);
       for (R_xlen_t j = 0; j < m; ++k, ++j){
-        SET_VECTOR_ELT(out, k, p_temp[j]);
+        SET_VECTOR_ELT(out, k, p_vec[j]);
       }
     }
     break;
   }
+  case r_fct: {
 
-  default: {
-    SEXP c_char = SHIELD(make_utf8_str("c")); ++NP;
-    out = SHIELD(base_do_call(c_char, x)); ++NP;
+    SHIELD(out = cpp_rep_len(vec_template, out_size)); ++NP;
+
+    int* RESTRICT p_out = INTEGER(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_factor_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(INTEGER_RO(vec), m, &p_out[k]);
+    }
     break;
   }
-  }
-  if (is_date2){
+  case r_date: {
+
+    if (TYPEOF(vec_template) == INTSXP){
+    SHIELD(out = init<r_integer_t>(out_size, false)); ++NP;
     Rf_classgets(out, make_utf8_str("Date"));
+
+    int* RESTRICT p_out = INTEGER(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_date_t>(p_x[i], vec_template), vec_idx);
+      R_Reprotect(vec = coerce_vec(vec, INTSXP), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(INTEGER_RO(vec), m, &p_out[k]);
+    }
+  } else {
+    SHIELD(out = init<r_numeric_t>(out_size, false)); ++NP;
+    Rf_classgets(out, make_utf8_str("Date"));
+
+    double* RESTRICT p_out = REAL(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_date_t>(p_x[i], vec_template), vec_idx);
+      R_Reprotect(vec = coerce_vec(vec, REALSXP), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(REAL_RO(vec), m, &p_out[k]);
+    }
   }
-  if (is_datetime2){
-    Rf_copyMostAttrib(p_x[first_datetime], out);
+  break;
   }
+  case r_pxct: {
+
+    SHIELD(out = init<r_posixt_t>(out_size, false)); ++NP;
+    SEXP out_tzone = SHIELD(Rf_getAttrib(vec_template, install_utf8("tzone"))); ++NP;
+    Rf_setAttrib(out, install_utf8("tzone"), out_tzone);
+
+    double* RESTRICT p_out = REAL(out);
+
+    for (int i = 0; i < n; ++i, k += m){
+      R_Reprotect(vec = cast<r_posixt_t>(p_x[i], vec_template), vec_idx);
+      m = Rf_xlength(vec);
+      std::copy_n(REAL_RO(vec), m, &p_out[k]);
+    }
+    break;
+  }
+  case r_df: {
+    SHIELD(out = cpp_df_c(x)); ++NP;
+    break;
+  }
+  case r_unk: {
+    SEXP call = SHIELD(coerce_vec(x, LISTSXP)); ++NP;
+    SHIELD(call = Rf_lcons(install_utf8("c"), call)); ++NP;
+    SHIELD(out = Rf_eval(call, R_GetCurrentEnv())); ++NP;
+    break;
+  }
+  default: {
+    YIELD(NP);
+    Rf_error("Don't know how to combine elements");
+  }
+  }
+
+  if (has_top_level_names){
+    SEXP top_level_names = SHIELD(get_vec_names(x)); ++NP;
+    SEXP name_sizes = SHIELD(cpp_lengths(x, false)); ++NP;
+    SHIELD(combined_names = cpp_rep(top_level_names, name_sizes)); ++NP;
+  }
+  SHIELD(out = set_vec_names(out, combined_names)); ++NP;
   YIELD(NP);
   return out;
 }
+
+// `c()` but no concatenation of names
+
+[[cpp11::register]]
+SEXP cpp_c(SEXP x){
+
+  if (!Rf_isVectorList(x)){
+    Rf_error("`x` must be a list of vectors");
+  }
+
+  int n = Rf_length(x);
+
+  // Cast all objects to common type
+  const SEXP *p_x = LIST_PTR_RO(x);
+
+  // Figure out final size
+  R_xlen_t out_size = 0;
+  for (int i = 0; i < n; ++i) out_size += vector_length(p_x[i]);
+
+  // 'vec_template' here acts as a template for the final result
+  SEXP vec_template = SHIELD(cpp_common_template(x));
+  SEXP out = SHIELD(combine_internal(x, out_size, vec_template));
+  YIELD(2);
+  return out;
+}
